/*
** Module   : ScrollBarGuiImpl.java
** Abstract : scroll bar implementation. 
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140917 First version.
** 002 MAG 20141013 Fixed system colors. Handle mouse events.
** 003 HC  20141021 Implemented window move/position support.
** 004 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 005 CA  20141123 Fixed scrollbar button dimensions.
** 006 CA  20150323 The GuiFontResolver needs to know if the expected font is for a fixed-width
**                  widget.
** 007 CA  20150509 Changed to allow double-value step for scrolling.
** 008 EVL 20150511 Consider scrollpane box when computing the scrollbar coordinates.  Scrollbar
**                  thumb size/position changing approach.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.event.*;
import java.util.concurrent.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * A GUI scroll bar implementation. 
 * TODO 
 * REPEAT_MILLIS constant is used to simulate mouse press repeated events. 
 * This parameter should be configurable for different customers and users.
 */
@SuppressWarnings("unchecked")
public class ScrollBarGuiImpl
extends ScrollBar<GuiOutputManager>
{
   /** Thumb button minimum size */
   private static final int THUMB_SIZE = 8;
   
   /** Minimum thumb size to draw stripes */
   private static final int STRIPES_SIZE = 16;

   /** Screen driver */
   private GuiDriver gd;
   
   /** Color resolver */
   private GuiColorResolver gc;
      
   /** Left button */
   private ScrollBarGuiButton lButt;
   
   /** Right button */
   private ScrollBarGuiButton rButt;
   
   /** Thumb button position */
   private NativePoint thumbPos;
   
   /** Thumb button size */
   private NativeDimension thumbDim;
   
   /** Width correction */
   double wOff = 0;
   
   /** Height correction */
   double hOff = 0;
   
   /** Mouse pressed indicator */
   volatile boolean pressed;
   
   /**
    * Default constructor.
    * 
    * @param   id
    *          The widget's ID.
    */
   public ScrollBarGuiImpl(WidgetId id)
   {
      super(id);
      init();
   }

   /**
    * Constructor.
    * 
    * @param   id
    *          The widget's ID.
    * @param   orientation
    *          Scroll bar orientation.
    */
   ScrollBarGuiImpl(WidgetId id, Orientation orientation)
   {
      super(id, orientation);
      init();
   }
   
   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#width()
    */
   @Override
   public double width()
   {
      boolean parentWidth = (isInheritWidth() && parent() != null);
      return (parentWidth ? parent().width() : super.width()) - wOff;
   }
   
   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#height()
    */
   @Override
   public double height()
   {
      boolean parentHeight = (isInheritHeight() && parent() != null);
      return (parentHeight ? parent().height() : super.height()) - hOff;
   }
   
   /**
    * Widget drawing method.
    */
   @Override
   public void draw()
   {
      // draw only when widget is visible
      if (!isVisible())
      {
         return;
      }

      gc.colors();
      
      // get scroll bar coordinates
      final NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      
      final NativeRectangle clip = clipRectangle(p, d);
            
      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            // draw scroll bar area
            gd.setColor(gc.bgColor);
            gd.fillRect(0, 0, d.width + 1, d.height + 1);

            // draw buttons
            drawScrollBarButtons();                                    
         }
         
         /**
          * Draw scroll bar buttons.  Layout buttons and call draw method
          */
         private void drawScrollBarButtons()
         {
            int h, w;
            
            if (orientation == Orientation.HORIZONTAL)
            {
               // compute buttons size
               h = GuiOptions.SCROLLBAR_SIZE;
               w = Math.min(d.width / 2, GuiOptions.SCROLLBAR_SIZE);
            }   
            else
            {
               // compute buttons size
               w = GuiOptions.SCROLLBAR_SIZE;
               h = Math.min(d.height / 2, GuiOptions.SCROLLBAR_SIZE);
            }            
            
            // draw buttons
            drawScrollBarButtons(w, h);
            
            // draw scroll bar thumb only if need
            double range = range() + step;
            
            if (range > 0 && w == h)
            {
               drawScrollBarThumb(w, h, range);
            }
         }

         /**
          * Draw scroll bar buttons.
          * 
          * @param   w
          *          Scroll bar buttons width.
          * @param   h
          *          Scroll bar buttons height.
          */
         private void drawScrollBarButtons(int w, int h)
         {
            // left (top) button
            lButt.config().widthPixels = w;
            lButt.config().heightPixels = h;
            lButt.setPhysicalLocation(0, 0);
            lButt.setLocation(0, 0);
            lButt.draw();
            
            // right (bottom) button
            rButt.config().widthPixels = w;
            rButt.config().heightPixels = h;
            NativePoint np = null;            
            if (orientation == Orientation.HORIZONTAL)
            {
               np = new NativePoint(d.width - w, 0);
            }
            else
            {
               np = new NativePoint(0, d.height - h);
            }
            rButt.setPhysicalLocation(np.x, np.y);    
            rButt.setLocation(screen().coordinates().pointFromNative(np));
            rButt.draw();            
         }

         /**
          * Draw scroll bar thumb.
          * 
          * @param   w
          *          Scroll bar buttons width.
          * @param   h
          *          Scroll bar buttons height.
          * @param   range
          *          Scroll bar range  
          */
         private void drawScrollBarThumb(int w, int h, double range)
         {
            
            if (orientation == Orientation.HORIZONTAL)
            {               
               int delta = d.width - 2 * w;
               double unit = Coordinate.toInt(delta * visibleRange / range); 
                              
               thumbPos.x = w + Coordinate.toInt(unit * position / range);
               thumbPos.y = 0;
               thumbDim.width = Coordinate.toInt(unit);
               thumbDim.height = h;

               // if thumb size is small use default size
               if (thumbDim.width < THUMB_SIZE)
               {
                  int test = delta - THUMB_SIZE;
                  if (test > 0)
                  {
                     unit = test / range;
                     thumbPos.x = w + Coordinate.toInt(unit * position);
                     thumbDim.width = THUMB_SIZE;
                  }
               }

               // draw correction at the right
               int maxWidth = delta + w;
               if ((position == max) || ((thumbPos.x + thumbDim.width) > maxWidth))
               {
                  thumbPos.x = maxWidth - thumbDim.width;
               }

               gd.setColor(gc.thumbColor);
               gd.fill3DRect(thumbPos.x, thumbPos.y, thumbDim.width, thumbDim.height, true);
               
               // stripes
               if (thumbDim.width > STRIPES_SIZE)
               {
                  // gap between stripes
                  int g = 4;                  
                  int x = thumbDim.width / 2;
                  int y = thumbDim.height - 2 * g;
                  gd.setColor(gc.bgColor);
                  gd.translate(thumbPos.x, 0);
                  gd.fill3DRect(x - g, g, 2, y, false);
                  gd.fill3DRect(x,     g, 2, y, false);
                  gd.fill3DRect(x + g, g, 2, y, false);
                  gd.translate(-thumbPos.x, 0);
               }
            }   
            else
            {
               // draw scroll bar thumb
               int delta = d.height - 2 * h;
               double unit = Coordinate.toInt(delta * visibleRange / range); 
               
               thumbPos.x = 0;
               thumbPos.y = h + Coordinate.toInt(unit * position / range);
               thumbDim.width = w;
               thumbDim.height = Coordinate.toInt(unit); // thumb size
               
               // if thumb size is small use default size
               if (thumbDim.height < THUMB_SIZE)
               {
                  int test = delta - THUMB_SIZE;
                  if (test > 0)
                  {
                     unit = test / range;
                     thumbPos.y = h + Coordinate.toInt(unit * position);
                     thumbDim.height = THUMB_SIZE;
                  }
               }

               // draw correction at the bottom
               int maxHeight = delta + h;
               if ((position == max) || ((thumbPos.y + thumbDim.height) > maxHeight))
               {
                  thumbPos.y = maxHeight - thumbDim.height;
               }
               
               gd.setColor(gc.thumbColor);
               gd.fill3DRect(thumbPos.x, thumbPos.y, thumbDim.width, thumbDim.height, true);

               // stripes
               if (thumbDim.height > STRIPES_SIZE)
               {
                  // gap between stripes
                  int g = 4;               
                  int y = thumbDim.height / 2;
                  int x = thumbDim.width - 2 * g;
                  gd.setColor(gc.bgColor);
                  gd.translate(0, thumbPos.y);
                  gd.fill3DRect(g, y - g, x, 2, false);
                  gd.fill3DRect(g, y,     x, 2, false);
                  gd.fill3DRect(g, y + g, x, 2, false);
                  gd.translate(0, -thumbPos.y);
               }
            }
         }            
      });
   }

   /**
    * Attach this widget to the specified container.
    * 
    * @param    parent
    *           The parent container.
    */
   @Override
   public void attach(AbstractContainer<GuiOutputManager> parent)
   {
      if (parent instanceof ScrollListener)
      {
         addScrollListener((ScrollListener) parent);
      }
   }

   /**
    * Layout widget.  Before any draw operation adjust widget physical location.
    */
   @Override
   public NativePoint physicalLocation()
   {
      Container<GuiOutputManager> parent = parent();
      
      NativeDimension d2 = parent.physicalDimension();
      NativeDimension d1 = physicalDimension();
      
      // if parent has the 1-pixel frame - need to shift accordingly
      boolean parentBoxed = isParentBoxed(parent);

      int x = 0;
      int y = 0;
      
      if (orientation == Orientation.HORIZONTAL)
      {
         x = parentBoxed ? x + 1 : x;
         y = parentBoxed ? d2.height - d1.height - 2 : d2.height - d1.height;
      }
      else
      {
         x = parentBoxed ? d2.width - d1.width - 2 : d2.width - d1.width;
         y = parentBoxed ? y + 1 : y;
      }      
      NativePoint np = new NativePoint(x, y);
      setLocation(screen().coordinates().pointFromNative(np));
      setPhysicalLocation(np.x, np.y);
      return super.physicalLocation();
   }
   
   /**
    * Layout widget.  Before any draw operation adjust widget physical location.
    */
   @Override
   public NativeDimension physicalDimension()
   {
      NativeDimension dimResults = super.physicalDimension();

      // check if the paren has one pixel border      
      boolean parentBoxed = isParentBoxed(parent());

      // if parent s boxed - need to decrease dimension by two pixels
      if (orientation == Orientation.HORIZONTAL)
      {
         dimResults.width = parentBoxed ? dimResults.width - 2 : dimResults.width;   
      }
      else
      {
         dimResults.height = parentBoxed ? dimResults.height - 2 : dimResults.height;   
      }      
      
      return dimResults;
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      synchronized (e)
      {
         if (!isOverThumbButton(e))
         {
            Position direction = scrollDirection(e);
            onButtonClick(direction);
            refresh();
         }
      }
   }
      
   /**
    * Notification of a mouse dragged event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      pressed = false;
      
      synchronized (e)
      {
         Position direction = scrollDirection(e);
         double pos  = position(e);
         onDrag(direction, pos);
         refresh();
      }
   }
   
   /**
    * Notification of a mouse pressed event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(final MouseEvent e)
   {
      pressed = true;
      if (!isOverThumbButton(e))
      {
         ExecutorService executor = Executors.newSingleThreadExecutor();
         executor.execute(new Runnable() {
            
            @Override
            public void run()
            {
               while (pressed)
               {
                  try
                  {
                     // wait to see if not a click event
                     Thread.sleep(GuiOptions.REPEAT_MILLIS);
                     // fire mouse clicked
                     if (pressed)
                     {
                        mouseClicked(e);
                     }
                  } 
                  catch (InterruptedException ex)
                  {
                     // nop
                  }
               }
            }
            
         });
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */   
   @Override
   public void mouseReleased(MouseEvent e)
   {
      pressed = false;
   }

   /**
    * Notification of a mouse wheel moved event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */   
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      Position direction;
      int rotation = e.getWheelRotation();

      // determine scroll direction
      if (orientation == Orientation.HORIZONTAL)
      {
         direction = (rotation < 0) ? Position.LEFT : Position.RIGHT;
      }
      else
      {
         direction = (rotation < 0) ? Position.LEFT : Position.RIGHT;
      }      
      
      // do scroll
      for (int i = 0;i < Math.abs(rotation);i++)
      {
         onButtonClick(direction);
         refresh();
      }
   }

   /**
    * Scroll bar initialization.
    */
   private void init()
   {
      gd = (GuiDriver) OutputManager.getDriver();
      // set default BGCOLOR and FGCOLOR
      gc = new GuiColorResolver(gd, config(), SysColor.COLOR_SCROLLBAR, SysColor.COLOR_WINDOWTEXT);
      gc.thumbColor = gd.getSysColor(SysColor.COLOR_3DFACE);
      // buttons
      thumbPos = new NativePoint(0, 0);
      thumbDim = new NativeDimension(0, 0);
      lButt = new ScrollBarGuiButton(WidgetId.nextID(), this, ScrollBar.Position.LEFT);
      rButt = new ScrollBarGuiButton(WidgetId.nextID(), this, ScrollBar.Position.RIGHT);
      add(lButt);
      add(rButt);
      
      Window<?> window = (Window<?>) ancestor();
      GuiFontResolver gf = new GuiFontResolver(window, gd, config(), false);
      
      FontCache cached = gf.font();
      CoordinatesConversion cc = screen().coordinates();
      double dim = cc.widthFromNative(cached.fontHeight);
      
      if (orientation == Orientation.HORIZONTAL)
      {
         dim = cc.heightFromNative(cc.widthToNative(dim));
         setHeight(dim);
      }
      else
      {
         setWidth(dim);
      }
   }      
   
   /**
    * Check if mouse is over thumb button.
    * 
    * @param   event
    *          Mouse event.
    *          
    * @return  <code>true</code> if mouse under the thumb button <code>false</code> otherwise. 
    */
   private boolean isOverThumbButton(MouseEvent event)
   {
      NativeRectangle thumbRect = new NativeRectangle(thumbPos, thumbDim);
      NativePoint mousePos = translate(this, event);
      return thumbRect.contains(mousePos);       
   }

   /**
    * Get box flag indicating drawing frame around the parent scrollpane.
    * 
    * @param   parent
    *          Parent container to check.
    * 
    * @return  see above
    */
   private boolean isParentBoxed(Container<GuiOutputManager> parent)
   {
      return parent instanceof ScrollPaneGuiImpl && ((ScrollPaneGuiImpl)parent).getBoxFlag();
   }
   
   /**
    * Determine the equivalent scroll bar position which correspond to actual mouse position. 
    * 
    * @param   e
    *          Mouse event.
    * 
    * @return  see above.
    */
   public double position(MouseEvent e)
   {
      NativePoint mousePos = translate(this, e);
      NativeDimension d = physicalDimension();

      if (orientation == Orientation.HORIZONTAL)
      {
         NativeDimension b = lButt.physicalDimension();
         int domain = d.width - b.width - thumbDim.width;
         return domain == 0 ? 0 : (mousePos.x * max / domain);
      }
      else
      {
         NativeDimension b = rButt.physicalDimension();
         int domain = d.height - b.height - thumbDim.height;
         return domain == 0 ? 0 : (mousePos.y * max / domain);
      }
   }
   
   /**
    * Determine scroll direction 
    * 
    * @param   e
    *          Mouse event.
    *          
    * @return  see above.
    */
   private Position scrollDirection(MouseEvent e)
   {
      NativePoint n = translate(this, e);
      NativeDimension d = this.physicalDimension();
      
      if (orientation == Orientation.HORIZONTAL)
      {
         return (n.x < d.width / 2) ? Position.LEFT : Position.RIGHT;
      }
      else
      {
         return (n.y < d.height / 2) ? Position.LEFT : Position.RIGHT;
      }      
   }
   
   /**
    * Invoke repaint for this widget.
    */
   private void refresh()
   {
      ThinClient.getInstance().eventDrawingBracket(this, false, true, new Runnable()
      {
         @Override
         public void run()
         {
            repaint();
         }
      });      
   }
}
